/* ANSI-C code produced by gperf version 3.3 */
/* Command-line: gperf --output-file src/nspawn/nspawn-gperf.c ../systemd/src/nspawn/nspawn-gperf.gperf  */
/* Computed positions: -k'12,$' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

#line 2 "../systemd/src/nspawn/nspawn-gperf.gperf"

_Pragma("GCC diagnostic ignored \"-Wimplicit-fallthrough\"")
#if __GNUC__ >= 15
_Pragma("GCC diagnostic ignored \"-Wzero-as-null-pointer-constant\"")
#endif
#include <stddef.h>
#include "conf-parser.h"
#include "nspawn-settings.h"
#include <string.h>

#define TOTAL_KEYWORDS 65
#define MIN_WORD_LENGTH 9
#define MAX_WORD_LENGTH 28
#define MIN_HASH_VALUE 9
#define MAX_HASH_VALUE 90
/* maximum key range = 82, duplicates = 0 */

#ifdef __GNUC__
__inline
#else
#ifdef __cplusplus
inline
#endif
#endif
static unsigned int
nspawn_gperf_hash (register const char *str, register size_t len)
{
  static const unsigned char asso_values[] =
    {
      91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
      91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
      91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
      91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
      91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
      91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
      91, 91, 91, 91, 91, 35, 91, 15, 60, 15,
      91, 25, 91, 45, 91, 45,  0, 91,  0, 35,
      15, 91, 40, 25, 20, 45,  0, 91, 91, 91,
      91, 91, 91, 91, 91, 91, 91, 20, 91, 40,
      50,  5,  0, 40,  0,  5, 91, 91, 15, 50,
      15, 40, 20, 91, 10, 10,  0,  5,  0, 91,
      91, 20, 91, 91, 91, 91, 91, 91, 91, 91,
      91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
      91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
      91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
      91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
      91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
      91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
      91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
      91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
      91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
      91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
      91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
      91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
      91, 91, 91, 91, 91, 91
    };
  register unsigned int hval = len;

  switch (hval)
    {
      default:
        hval += asso_values[(unsigned char)str[11]];
#if (defined __cplusplus && (__cplusplus >= 201703L || (__cplusplus >= 201103L && defined __clang__ && __clang_major__ + (__clang_minor__ >= 9) > 3))) || (defined __STDC_VERSION__ && __STDC_VERSION__ >= 202000L && ((defined __GNUC__ && __GNUC__ >= 10) || (defined __clang__ && __clang_major__ >= 9)))
      [[fallthrough]];
#elif (defined __GNUC__ && __GNUC__ >= 7) || (defined __clang__ && __clang_major__ >= 10)
      __attribute__ ((__fallthrough__));
#endif
      /*FALLTHROUGH*/
      case 11:
      case 10:
      case 9:
        break;
    }
  return hval + asso_values[(unsigned char)str[len - 1]];
}

const struct ConfigPerfItem *
nspawn_gperf_lookup (register const char *str, register size_t len)
{
#if (defined __GNUC__ && __GNUC__ + (__GNUC_MINOR__ >= 6) > 4) || (defined __clang__ && __clang_major__ >= 3)
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif
  static const struct ConfigPerfItem wordlist[] =
    {
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
      {(char*)0},
#line 22 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.Boot",                    config_parse_boot,                      0,                        0},
      {(char*)0}, {(char*)0},
#line 86 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Network.Port",                 config_parse_expose_port,               0,                        0},
      {(char*)0},
#line 81 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Network.IPVLAN",               config_parse_ipvlan_iface_pair,         0,                        offsetof(Settings, network_ipvlan)},
#line 80 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Network.MACVLAN",              config_parse_macvlan_iface_pair,        0,                        offsetof(Settings, network_macvlan)},
      {(char*)0}, {(char*)0}, {(char*)0},
#line 27 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.User",                    config_parse_string,                    CONFIG_PARSE_STRING_SAFE, offsetof(Settings, user)},
#line 77 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Network.Private",              config_parse_tristate,                  0,                        offsetof(Settings, private_network)},
      {(char*)0},
#line 85 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Network.Zone",                 config_parse_network_zone,              0,                        0},
#line 82 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Network.VirtualEthernet",      config_parse_tristate,                  0,                        offsetof(Settings, network_veth)},
#line 66 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Files.Volatile",               config_parse_volatile_mode,             0,                        offsetof(Settings, volatile_mode)},
#line 25 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.Parameters",              config_parse_strv,                      0,                        offsetof(Settings, parameters)},
#line 78 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Network.NamespacePath",        config_parse_path,                      0,                        offsetof(Settings, network_namespace_path)},
#line 79 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Network.Interface",            config_parse_network_iface_pair,        0,                        offsetof(Settings, network_interfaces)},
#line 70 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Files.Inaccessible",           config_parse_inaccessible,              0,                        0},
#line 58 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.OOMScoreAdjust",          config_parse_oom_score_adjust,          0,                        0},
#line 60 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.ResolvConf",              config_parse_resolv_conf,               0,                        offsetof(Settings, resolv_conf)},
#line 26 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.Environment",             config_parse_strv,                      0,                        offsetof(Settings, environment)},
#line 36 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.PrivateUsers",            config_parse_private_users,             0,                        0},
#line 62 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.Timezone",                config_parse_timezone_mode,             0,                        offsetof(Settings, timezone)},
#line 75 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Files.BindUser",               config_parse_bind_user,                 0,                        offsetof(Settings, bind_user)},
#line 37 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.PrivateUsersDelegate",    config_parse_unsigned,                  0,                        offsetof(Settings, delegate_container_ranges)},
#line 61 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.LinkJournal",             config_parse_link_journal,              0,                        0},
      {(char*)0},
#line 73 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Files.PrivateUsersChown",      config_parse_userns_chown,              0,                        offsetof(Settings, userns_ownership)},
#line 23 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.Ephemeral",               config_parse_tristate,                  0,                        offsetof(Settings, ephemeral)},
#line 57 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.NoNewPrivileges",         config_parse_tristate,                  0,                        offsetof(Settings, no_new_privileges)},
#line 59 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.CPUAffinity",             config_parse_cpu_set,                   0,                        offsetof(Settings, cpu_set)},
#line 29 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.AmbientCapability",       config_parse_capability,                0,                        offsetof(Settings, ambient_capability)},
#line 68 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Files.BindReadOnly",           config_parse_bind,                      1,                        0},
#line 76 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Files.BindUserShell",          config_parse_bind_user_shell,           0,                        0},
#line 48 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.LimitNPROC",              config_parse_rlimit,                    RLIMIT_NPROC,             offsetof(Settings, rlimit)},
#line 39 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.SystemCallFilter",        config_parse_syscall_filter,            0,                        0},
#line 74 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Files.PrivateUsersOwnership",  config_parse_userns_ownership,          0,                        offsetof(Settings, userns_ownership)},
#line 83 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Network.VirtualEthernetExtra", config_parse_veth_extra,                0,                        0},
#line 65 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Files.ReadOnly",               config_parse_tristate,                  0,                        offsetof(Settings, read_only)},
#line 28 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.Capability",              config_parse_capability,                0,                        offsetof(Settings, capability)},
#line 55 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.LimitRTTIME",             config_parse_rlimit,                    RLIMIT_RTTIME,            offsetof(Settings, rlimit)},
      {(char*)0},
#line 71 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Files.Overlay",                config_parse_overlay,                   0,                        0},
#line 35 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.PivotRoot",               config_parse_pivot_root,                0,                        0},
#line 41 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.LimitFSIZE",              config_parse_rlimit,                    RLIMIT_FSIZE,             offsetof(Settings, rlimit)},
#line 32 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.Personality",             config_parse_personality,               0,                        offsetof(Settings, personality)},
      {(char*)0},
#line 52 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.LimitMSGQUEUE",           config_parse_rlimit,                    RLIMIT_MSGQUEUE,          offsetof(Settings, rlimit)},
#line 30 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.DropCapability",          config_parse_capability,                0,                        offsetof(Settings, drop_capability)},
#line 67 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Files.Bind",                   config_parse_bind,                      0,                        0},
#line 72 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Files.OverlayReadOnly",        config_parse_overlay,                   1,                        0},
#line 47 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.LimitAS",                 config_parse_rlimit,                    RLIMIT_AS,                offsetof(Settings, rlimit)},
#line 45 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.LimitRSS",                config_parse_rlimit,                    RLIMIT_RSS,               offsetof(Settings, rlimit)},
#line 44 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.LimitCORE",               config_parse_rlimit,                    RLIMIT_CORE,              offsetof(Settings, rlimit)},
#line 24 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.ProcessTwo",              config_parse_pid2,                      0,                        0},
#line 46 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.LimitNOFILE",             config_parse_rlimit,                    RLIMIT_NOFILE,            offsetof(Settings, rlimit)},
#line 63 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.SuppressSync",            config_parse_tristate,                  0,                        offsetof(Settings, suppress_sync)},
#line 56 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.Hostname",                config_parse_hostname,                  0,                        offsetof(Settings, hostname)},
#line 84 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Network.Bridge",               config_parse_ifname,                    0,                        offsetof(Settings, network_bridge)},
#line 31 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.KillSignal",              config_parse_signal,                    0,                        offsetof(Settings, kill_signal)},
#line 54 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.LimitRTPRIO",             config_parse_rlimit,                    RLIMIT_RTPRIO,            offsetof(Settings, rlimit)},
      {(char*)0},
#line 40 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.LimitCPU",                config_parse_rlimit,                    RLIMIT_CPU,               offsetof(Settings, rlimit)},
#line 53 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.LimitNICE",               config_parse_rlimit,                    RLIMIT_NICE,              offsetof(Settings, rlimit)},
#line 50 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.LimitLOCKS",              config_parse_rlimit,                    RLIMIT_LOCKS,             offsetof(Settings, rlimit)},
#line 38 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.NotifyReady",             config_parse_tristate,                  0,                        offsetof(Settings, notify_ready)},
#line 49 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.LimitMEMLOCK",            config_parse_rlimit,                    RLIMIT_MEMLOCK,           offsetof(Settings, rlimit)},
#line 64 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.RestrictAddressFamilies", config_parse_restrict_address_families, 0,                        0},
#line 33 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.MachineID",               config_parse_id128,                     0,                        offsetof(Settings, machine_id)},
#line 43 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.LimitSTACK",              config_parse_rlimit,                    RLIMIT_STACK,             offsetof(Settings, rlimit)},
#line 34 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.WorkingDirectory",        config_parse_path,                      0,                        offsetof(Settings, working_directory)},
      {(char*)0}, {(char*)0},
#line 42 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.LimitDATA",               config_parse_rlimit,                    RLIMIT_DATA,              offsetof(Settings, rlimit)},
#line 69 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Files.TemporaryFileSystem",    config_parse_tmpfs,                     0,                        0},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
#line 51 "../systemd/src/nspawn/nspawn-gperf.gperf"
      {"Exec.LimitSIGPENDING",         config_parse_rlimit,                    RLIMIT_SIGPENDING,        offsetof(Settings, rlimit)}
    };
#if (defined __GNUC__ && __GNUC__ + (__GNUC_MINOR__ >= 6) > 4) || (defined __clang__ && __clang_major__ >= 3)
#pragma GCC diagnostic pop
#endif

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      register unsigned int key = nspawn_gperf_hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          register const char *s = wordlist[key].section_and_lvalue;

          if (s && *str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return (struct ConfigPerfItem *) 0;
}
