/* ANSI-C code produced by gperf version 3.3 */
/* Command-line: gperf --output-file src/network/networkd-gperf.c ../systemd/src/network/networkd-gperf.gperf  */
/* Computed positions: -k'6,12' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

#line 2 "../systemd/src/network/networkd-gperf.gperf"

_Pragma("GCC diagnostic ignored \"-Wimplicit-fallthrough\"")
#if __GNUC__ >= 15
_Pragma("GCC diagnostic ignored \"-Wzero-as-null-pointer-constant\"")
#endif
#include <stddef.h>
#include "conf-parser.h"
#include "networkd-address-label.h"
#include "networkd-conf.h"
#include "networkd-dhcp-common.h"
#include "networkd-dns.h"
#include "networkd-manager.h"
#include "networkd-route-util.h"
#include <string.h>

#define TOTAL_KEYWORDS 28
#define MIN_WORD_LENGTH 13
#define MAX_WORD_LENGTH 39
#define MIN_HASH_VALUE 15
#define MAX_HASH_VALUE 63
/* maximum key range = 49, duplicates = 0 */

#ifdef __GNUC__
__inline
#else
#ifdef __cplusplus
inline
#endif
#endif
static unsigned int
networkd_gperf_hash (register const char *str, register size_t len)
{
  static const unsigned char asso_values[] =
    {
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64,  0, 64, 20, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 20, 64, 64, 30, 64,
      64, 64, 64, 64, 64, 64, 15, 64, 64, 64,
       0, 64, 15, 64,  0, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64,  5, 64, 20,
      10,  0, 64, 64, 64, 64, 64, 64, 64, 64,
       3,  0,  0, 64, 10, 64, 10, 64,  0,  0,
       0, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
      64, 64, 64, 64, 64, 64
    };
  return len + asso_values[(unsigned char)str[11]] + asso_values[(unsigned char)str[5]];
}

const struct ConfigPerfItem *
networkd_gperf_lookup (register const char *str, register size_t len)
{
#if (defined __GNUC__ && __GNUC__ + (__GNUC_MINOR__ >= 6) > 4) || (defined __clang__ && __clang_major__ >= 3)
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif
  static const struct ConfigPerfItem wordlist[] =
    {
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
      {(char*)0}, {(char*)0}, {(char*)0},
#line 42 "../systemd/src/network/networkd-gperf.gperf"
      {"DHCPv4.DUIDType",                         config_parse_duid_type,                  0,                                                         offsetof(Manager, dhcp_duid)},
#line 52 "../systemd/src/network/networkd-gperf.gperf"
      {"/* Deprecated */"},
#line 40 "../systemd/src/network/networkd-gperf.gperf"
      {"DHCPv4.UseDomains",                       config_parse_use_domains,                0,                                                         offsetof(Manager, dhcp_use_domains)},
#line 50 "../systemd/src/network/networkd-gperf.gperf"
      {"DHCPRelay.RemoteId",                      config_parse_dhcp_option,                /* check_length= */ true,                                  offsetof(Manager, dhcp_relay_remote_id)},
      {(char*)0}, {(char*)0},
#line 51 "../systemd/src/network/networkd-gperf.gperf"
      {"DHCPRelay.ExtraOption",                   config_parse_dhcp_option_tlv,            0,                                                         offsetof(Manager, dhcp_relay_extra_options)},
      {(char*)0},
#line 48 "../systemd/src/network/networkd-gperf.gperf"
      {"DHCPRelay.ServerAddress",                 config_parse_in_addr_non_null,           AF_INET,                                                   offsetof(Manager, dhcp_relay_server_address)},
#line 47 "../systemd/src/network/networkd-gperf.gperf"
      {"DHCPServer.PersistLeases",                config_parse_dhcp_server_persist_leases, 0,                                                         offsetof(Manager, dhcp_server_persist_leases)},
      {(char*)0},
#line 41 "../systemd/src/network/networkd-gperf.gperf"
      {"DHCPv4.ClientIdentifier",                 config_parse_dhcp_client_identifier,     0,                                                         offsetof(Manager, dhcp_client_identifier)},
      {(char*)0},
#line 27 "../systemd/src/network/networkd-gperf.gperf"
      {"Network.SpeedMeter",                      config_parse_bool,                       0,                                                         offsetof(Manager, use_speed_meter)},
      {(char*)0}, {(char*)0}, {(char*)0},
#line 33 "../systemd/src/network/networkd-gperf.gperf"
      {"Network.IPv4Forwarding",                  config_parse_tristate,                   0,                                                         offsetof(Manager, ip_forwarding[0])},
#line 43 "../systemd/src/network/networkd-gperf.gperf"
      {"DHCPv4.DUIDRawData",                      config_parse_duid_rawdata,               0,                                                         offsetof(Manager, dhcp_duid)},
#line 49 "../systemd/src/network/networkd-gperf.gperf"
      {"DHCPRelay.OverrideServerIdentifier",      config_parse_bool,                       0,                                                         offsetof(Manager, dhcp_relay_override_server_id)},
#line 45 "../systemd/src/network/networkd-gperf.gperf"
      {"DHCPv6.DUIDType",                         config_parse_duid_type,                  0,                                                         offsetof(Manager, dhcp6_duid)},
      {(char*)0},
#line 44 "../systemd/src/network/networkd-gperf.gperf"
      {"DHCPv6.UseDomains",                       config_parse_use_domains,                0,                                                         offsetof(Manager, dhcp6_use_domains)},
#line 32 "../systemd/src/network/networkd-gperf.gperf"
      {"Network.RouteTable",                      config_parse_route_table_names,          0,                                                         0},
#line 28 "../systemd/src/network/networkd-gperf.gperf"
      {"Network.SpeedMeterIntervalSec",           config_parse_sec,                        0,                                                         offsetof(Manager, speed_meter_interval_usec)},
      {(char*)0}, {(char*)0},
#line 30 "../systemd/src/network/networkd-gperf.gperf"
      {"Network.ManageForeignRoutes",             config_parse_bool,                       0,                                                         offsetof(Manager, manage_foreign_routes)},
#line 53 "../systemd/src/network/networkd-gperf.gperf"
      {"DHCP.DUIDType",                           config_parse_manager_duid_type,          0,                                                         0},
#line 31 "../systemd/src/network/networkd-gperf.gperf"
      {"Network.ManageForeignNextHops",           config_parse_bool,                       0,                                                         offsetof(Manager, manage_foreign_nexthops)},
      {(char*)0},
#line 54 "../systemd/src/network/networkd-gperf.gperf"
      {"DHCP.DUIDRawData",                        config_parse_manager_duid_rawdata,       0,                                                         0},
#line 39 "../systemd/src/network/networkd-gperf.gperf"
      {"IPv6AddressLabel.Label",                  config_parse_ipv6_address_label_section, IPV6_ADDRESS_LABEL_BY_MANAGER | IPV6_ADDRESS_LABEL,        0},
#line 38 "../systemd/src/network/networkd-gperf.gperf"
      {"IPv6AddressLabel.Prefix",                 config_parse_ipv6_address_label_section, IPV6_ADDRESS_LABEL_BY_MANAGER | IPV6_ADDRESS_LABEL_PREFIX, 0},
      {(char*)0}, {(char*)0}, {(char*)0},
#line 34 "../systemd/src/network/networkd-gperf.gperf"
      {"Network.IPv6Forwarding",                  config_parse_tristate,                   0,                                                         offsetof(Manager, ip_forwarding[1])},
#line 46 "../systemd/src/network/networkd-gperf.gperf"
      {"DHCPv6.DUIDRawData",                      config_parse_duid_rawdata,               0,                                                         offsetof(Manager, dhcp6_duid)},
#line 29 "../systemd/src/network/networkd-gperf.gperf"
      {"Network.ManageForeignRoutingPolicyRules", config_parse_bool,                       0,                                                         offsetof(Manager, manage_foreign_rules)},
      {(char*)0}, {(char*)0}, {(char*)0},
#line 36 "../systemd/src/network/networkd-gperf.gperf"
      {"Network.UseDomains",                      config_parse_use_domains,                0,                                                         offsetof(Manager, use_domains)},
#line 35 "../systemd/src/network/networkd-gperf.gperf"
      {"Network.IPv6PrivacyExtensions",           config_parse_ipv6_privacy_extensions,    0,                                                         offsetof(Manager, ipv6_privacy_extensions)},
      {(char*)0}, {(char*)0}, {(char*)0},
#line 37 "../systemd/src/network/networkd-gperf.gperf"
      {"IPv6AcceptRA.UseDomains",                 config_parse_use_domains,                0,                                                         offsetof(Manager, ndisc_use_domains)}
    };
#if (defined __GNUC__ && __GNUC__ + (__GNUC_MINOR__ >= 6) > 4) || (defined __clang__ && __clang_major__ >= 3)
#pragma GCC diagnostic pop
#endif

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      register unsigned int key = networkd_gperf_hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          register const char *s = wordlist[key].section_and_lvalue;

          if (s && *str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return (struct ConfigPerfItem *) 0;
}
