const char* statx_mask_one_to_name(unsigned mask) {
        switch (mask) {
        case STATX_SUBVOL: return "STATX_SUBVOL";
        case STATX_BTIME: return "STATX_BTIME";
        case STATX_MTIME: return "STATX_MTIME";
        case STATX_NLINK: return "STATX_NLINK";
        case STATX_UID: return "STATX_UID";
        case STATX_SIZE: return "STATX_SIZE";
        case STATX_DIOALIGN: return "STATX_DIOALIGN";
        case STATX_GID: return "STATX_GID";
        case STATX_DIO_READ_ALIGN: return "STATX_DIO_READ_ALIGN";
        case STATX_CTIME: return "STATX_CTIME";
        case STATX_MNT_ID_UNIQUE: return "STATX_MNT_ID_UNIQUE";
        case STATX_MODE: return "STATX_MODE";
        case STATX_ATIME: return "STATX_ATIME";
        case STATX_WRITE_ATOMIC: return "STATX_WRITE_ATOMIC";
        case STATX_MNT_ID: return "STATX_MNT_ID";
        case STATX_BLOCKS: return "STATX_BLOCKS";
        case STATX_INO: return "STATX_INO";
        case STATX_TYPE: return "STATX_TYPE";
        default: return NULL;
        }
}
