/* SPDX-License-Identifier: LGPL-2.1-or-later */
#pragma once

/* IWYU pragma: always_keep */

#include "shared-forward.h" /* IWYU pragma: export */

typedef struct Machine Machine;
typedef struct Manager Manager;
typedef struct Operation Operation;
