/* ANSI-C code produced by gperf version 3.3 */
/* Command-line: gperf --output-file src/login/logind-gperf.c ../systemd/src/login/logind-gperf.gperf  */
/* Computed positions: -k'7,15' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

#line 2 "../systemd/src/login/logind-gperf.gperf"

_Pragma("GCC diagnostic ignored \"-Wimplicit-fallthrough\"")
#if __GNUC__ >= 15
_Pragma("GCC diagnostic ignored \"-Wzero-as-null-pointer-constant\"")
#endif
#include <stddef.h>
#include "conf-parser.h"
#include "logind.h"
#include "logind-user.h"
#include <string.h>

#define TOTAL_KEYWORDS 37
#define MIN_WORD_LENGTH 14
#define MAX_WORD_LENGTH 34
#define MIN_HASH_VALUE 14
#define MAX_HASH_VALUE 71
/* maximum key range = 58, duplicates = 0 */

#ifdef __GNUC__
__inline
#else
#ifdef __cplusplus
inline
#endif
#endif
static unsigned int
logind_gperf_hash (register const char *str, register size_t len)
{
  static const unsigned char asso_values[] =
    {
      72, 72, 72, 72, 72, 72, 72, 72, 72, 72,
      72, 72, 72, 72, 72, 72, 72, 72, 72, 72,
      72, 72, 72, 72, 72, 72, 72, 72, 72, 72,
      72, 72, 72, 72, 72, 72, 72, 72, 72, 72,
      72, 72, 72, 72, 72, 72, 72, 72, 72, 72,
      72, 72, 72, 72, 72, 72, 72, 72, 72, 72,
      72, 72, 72, 72, 72, 72, 72, 20, 40, 72,
      72, 72, 15, 10, 72,  0,  0,  0,  0, 72,
       0, 72, 35,  0, 10,  0, 72,  5, 72, 72,
      72, 72, 72, 72, 72, 72, 72,  5,  5, 10,
       0,  0, 72, 72,  0,  0, 72, 72, 72, 72,
      72, 30, 72, 72,  0,  0, 72,  0, 72, 10,
      72,  0, 72, 72, 72, 72, 72, 72, 72, 72,
      72, 72, 72, 72, 72, 72, 72, 72, 72, 72,
      72, 72, 72, 72, 72, 72, 72, 72, 72, 72,
      72, 72, 72, 72, 72, 72, 72, 72, 72, 72,
      72, 72, 72, 72, 72, 72, 72, 72, 72, 72,
      72, 72, 72, 72, 72, 72, 72, 72, 72, 72,
      72, 72, 72, 72, 72, 72, 72, 72, 72, 72,
      72, 72, 72, 72, 72, 72, 72, 72, 72, 72,
      72, 72, 72, 72, 72, 72, 72, 72, 72, 72,
      72, 72, 72, 72, 72, 72, 72, 72, 72, 72,
      72, 72, 72, 72, 72, 72, 72, 72, 72, 72,
      72, 72, 72, 72, 72, 72, 72, 72, 72, 72,
      72, 72, 72, 72, 72, 72, 72, 72, 72, 72,
      72, 72, 72, 72, 72, 72
    };
  register unsigned int hval = len;

  switch (hval)
    {
      default:
        hval += asso_values[(unsigned char)str[14]];
#if (defined __cplusplus && (__cplusplus >= 201703L || (__cplusplus >= 201103L && defined __clang__ && __clang_major__ + (__clang_minor__ >= 9) > 3))) || (defined __STDC_VERSION__ && __STDC_VERSION__ >= 202000L && ((defined __GNUC__ && __GNUC__ >= 10) || (defined __clang__ && __clang_major__ >= 9)))
      [[fallthrough]];
#elif (defined __GNUC__ && __GNUC__ >= 7) || (defined __clang__ && __clang_major__ >= 10)
      __attribute__ ((__fallthrough__));
#endif
      /*FALLTHROUGH*/
      case 14:
      case 13:
      case 12:
      case 11:
      case 10:
      case 9:
      case 8:
      case 7:
        hval += asso_values[(unsigned char)str[6]];
        break;
    }
  return hval;
}

const struct ConfigPerfItem *
logind_gperf_lookup (register const char *str, register size_t len)
{
#if (defined __GNUC__ && __GNUC__ + (__GNUC_MINOR__ >= 6) > 4) || (defined __clang__ && __clang_major__ >= 3)
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif
  static const struct ConfigPerfItem wordlist[] =
    {
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
      {(char*)0}, {(char*)0}, {(char*)0}, {(char*)0},
      {(char*)0}, {(char*)0},
#line 23 "../systemd/src/login/logind-gperf.gperf"
      {"Login.NAutoVTs",                     config_parse_n_autovts,             0, offsetof(Manager, n_autovts)},
      {(char*)0}, {(char*)0},
#line 55 "../systemd/src/login/logind-gperf.gperf"
      {"Login.SessionsMax",                  config_parse_uint64,                0, offsetof(Manager, sessions_max)},
#line 57 "../systemd/src/login/logind-gperf.gperf"
      {"Login.UserTasksMax",                 config_parse_compat_user_tasks_max, 0, 0},
#line 26 "../systemd/src/login/logind-gperf.gperf"
      {"Login.KillOnlyUsers",                config_parse_strv,                  0, offsetof(Manager, kill_only_users)},
#line 30 "../systemd/src/login/logind-gperf.gperf"
      {"Login.SleepOperation",               config_parse_handle_action_sleep,   0, offsetof(Manager, handle_action_sleep_mask)},
      {(char*)0},
#line 27 "../systemd/src/login/logind-gperf.gperf"
      {"Login.KillExcludeUsers",             config_parse_strv,                  0, offsetof(Manager, kill_exclude_users)},
#line 25 "../systemd/src/login/logind-gperf.gperf"
      {"Login.KillUserProcesses",            config_parse_bool,                  0, offsetof(Manager, kill_user_processes)},
#line 58 "../systemd/src/login/logind-gperf.gperf"
      {"Login.StopIdleSessionSec",           config_parse_sec_fix_0,             0, offsetof(Manager, stop_idle_session_usec)},
      {(char*)0}, {(char*)0}, {(char*)0},
#line 59 "../systemd/src/login/logind-gperf.gperf"
      {"Login.WallMessages",                 config_parse_bool,                  0, offsetof(Manager, wall_messages)},
#line 54 "../systemd/src/login/logind-gperf.gperf"
      {"Login.InhibitorsMax",                config_parse_uint64,                0, offsetof(Manager, inhibitors_max)},
#line 46 "../systemd/src/login/logind-gperf.gperf"
      {"Login.LidSwitchIgnoreInhibited",     config_parse_bool,                  0, offsetof(Manager, lid_switch_ignore_inhibited)},
#line 44 "../systemd/src/login/logind-gperf.gperf"
      {"Login.SuspendKeyIgnoreInhibited",    config_parse_bool,                  0, offsetof(Manager, suspend_key_ignore_inhibited)},
      {(char*)0}, {(char*)0},
#line 28 "../systemd/src/login/logind-gperf.gperf"
      {"Login.InhibitDelayMaxSec",           config_parse_sec,                   0, offsetof(Manager, inhibit_delay_max)},
      {(char*)0},
#line 39 "../systemd/src/login/logind-gperf.gperf"
      {"Login.HandleLidSwitch",              config_parse_handle_action,         0, offsetof(Manager, handle_lid_switch)},
#line 35 "../systemd/src/login/logind-gperf.gperf"
      {"Login.HandleSuspendKey",             config_parse_handle_action,         0, offsetof(Manager, handle_suspend_key)},
#line 48 "../systemd/src/login/logind-gperf.gperf"
      {"Login.HoldoffTimeoutSec",            config_parse_sec,                   0, offsetof(Manager, holdoff_timeout_usec)},
#line 43 "../systemd/src/login/logind-gperf.gperf"
      {"Login.PowerKeyIgnoreInhibited",      config_parse_bool,                  0, offsetof(Manager, power_key_ignore_inhibited)},
      {(char*)0},
#line 33 "../systemd/src/login/logind-gperf.gperf"
      {"Login.HandleRebootKey",              config_parse_handle_action,         0, offsetof(Manager, handle_reboot_key)},
#line 41 "../systemd/src/login/logind-gperf.gperf"
      {"Login.HandleLidSwitchDocked",        config_parse_handle_action,         0, offsetof(Manager, handle_lid_switch_docked)},
      {(char*)0},
#line 37 "../systemd/src/login/logind-gperf.gperf"
      {"Login.HandleHibernateKey",           config_parse_handle_action,         0, offsetof(Manager, handle_hibernate_key)},
#line 31 "../systemd/src/login/logind-gperf.gperf"
      {"Login.HandlePowerKey",               config_parse_handle_action,         0, offsetof(Manager, handle_power_key)},
#line 36 "../systemd/src/login/logind-gperf.gperf"
      {"Login.HandleSuspendKeyLongPress",    config_parse_handle_action,         0, offsetof(Manager, handle_suspend_key_long_press)},
      {(char*)0},
#line 45 "../systemd/src/login/logind-gperf.gperf"
      {"Login.HibernateKeyIgnoreInhibited",  config_parse_bool,                  0, offsetof(Manager, hibernate_key_ignore_inhibited)},
#line 40 "../systemd/src/login/logind-gperf.gperf"
      {"Login.HandleLidSwitchExternalPower", config_parse_handle_action,         0, offsetof(Manager, handle_lid_switch_ep)},
#line 34 "../systemd/src/login/logind-gperf.gperf"
      {"Login.HandleRebootKeyLongPress",     config_parse_handle_action,         0, offsetof(Manager, handle_reboot_key_long_press)},
      {(char*)0}, {(char*)0},
#line 38 "../systemd/src/login/logind-gperf.gperf"
      {"Login.HandleHibernateKeyLongPress",  config_parse_handle_action,         0, offsetof(Manager, handle_hibernate_key_long_press)},
#line 32 "../systemd/src/login/logind-gperf.gperf"
      {"Login.HandlePowerKeyLongPress",      config_parse_handle_action,         0, offsetof(Manager, handle_power_key_long_press)},
#line 42 "../systemd/src/login/logind-gperf.gperf"
      {"Login.HandleSecureAttentionKey",     config_parse_handle_action,         0, offsetof(Manager, handle_secure_attention_key)},
#line 49 "../systemd/src/login/logind-gperf.gperf"
      {"Login.IdleAction",                   config_parse_handle_action,         0, offsetof(Manager, idle_action)},
      {(char*)0}, {(char*)0},
#line 50 "../systemd/src/login/logind-gperf.gperf"
      {"Login.IdleActionSec",                config_parse_sec,                   0, offsetof(Manager, idle_action_usec)},
#line 24 "../systemd/src/login/logind-gperf.gperf"
      {"Login.ReserveVT",                    config_parse_unsigned,              0, offsetof(Manager, reserve_vt)},
#line 51 "../systemd/src/login/logind-gperf.gperf"
      {"Login.RuntimeDirectorySize",         config_parse_tmpfs_size,            0, offsetof(Manager, runtime_dir_size)},
#line 29 "../systemd/src/login/logind-gperf.gperf"
      {"Login.UserStopDelaySec",             config_parse_sec,                   0, offsetof(Manager, user_stop_delay)},
      {(char*)0}, {(char*)0},
#line 47 "../systemd/src/login/logind-gperf.gperf"
      {"Login.RebootKeyIgnoreInhibited",     config_parse_bool,                  0, offsetof(Manager, reboot_key_ignore_inhibited)},
#line 52 "../systemd/src/login/logind-gperf.gperf"
      {"Login.RuntimeDirectoryInodesMax",    config_parse_iec_uint64,            0, offsetof(Manager, runtime_dir_inodes)},
      {(char*)0}, {(char*)0}, {(char*)0},
#line 53 "../systemd/src/login/logind-gperf.gperf"
      {"Login.RemoveIPC",                    config_parse_bool,                  0, offsetof(Manager, remove_ipc)},
#line 56 "../systemd/src/login/logind-gperf.gperf"
      {"Login.DesignatedMaintenanceTime",    config_parse_calendar,              0, offsetof(Manager, maintenance_time)}
    };
#if (defined __GNUC__ && __GNUC__ + (__GNUC_MINOR__ >= 6) > 4) || (defined __clang__ && __clang_major__ >= 3)
#pragma GCC diagnostic pop
#endif

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      register unsigned int key = logind_gperf_hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          register const char *s = wordlist[key].section_and_lvalue;

          if (s && *str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return (struct ConfigPerfItem *) 0;
}
